@@ -1,52 +1,13 @@
+#!/usr/bin/env python
+
 import os
 import requests
 import pandas as pd
 from time import sleep
 from tqdm import tqdm
+from .blockfrost_api_urls import *
 from typing import Union, Optional, List, Dict, Tuple
-
-# API Header parameter
-header_param_name = 'project_id'
-
-# Blockfrost Network URL
-bf_url_cardano_mainnet = 'https://cardano-mainnet.blockfrost.io/api/v0/'
-bf_url_cardano_testnet = 'https://cardano-testnet.blockfrost.io/api/v0/'
-
-# Blockfrost Stake URL
-bf_stake_url = 'accounts/'
-bf_stake_rewards_url = '/rewards'
-bf_stake_amount_history_url = '/history'
-bf_stake_delegation_url = '/delegations'
-bf_stake_registration_url = '/registrations'
-bf_stake_withdrawal_history_url = '/withdrawals'
-bf_stake_mir_history_url = '/mirs'
-bf_associated_addresses_url = '/addresses'
-bf_assets_associated_addresses_url = '/addresses/assets'
-
-# Blockfrost Addresse URL
-bf_address_url = 'addresses/'
-bf_address_details_url = '/total'
-bf_address_utxo_url = '/utxos'
-bf_address_transaction_url = '/transactions'
-
-# Blockfrost Network Information URL
-bf_network_informations_url = '/network'
-
-# Blockfrost Epochs URL
-bf_epoch_url = 'epochs/'
-bf_latest_epoch_url = 'epochs/latest'
-bf_latest_epoch_protocol_parameters_url = 'epochs/latest/parameters'
-
-# Blockfrost Pools URL
-bf_polls_url = 'pools/'
-bf_param_stake_pool_history_url = '/history'
-
-# Blockfrost Assets URL
-bf_assets_url = 'assets/'
-bf_asset_history_url = '/history'
-bf_asset_transactions_url = '/transactions'
-bf_asset_addresses_url = '/addresses'
-bf_assets_policy_url = 'policy/'
+from .utility import convert_hex_to_ascii, process_onchain_metadata, nb_results_to_return, set_query_string_parameter
 
 class Auth:
     def __init__(self, api_key: str=None, network: str="mainnet", proxies: dict=None):
@@ -59,7 +20,6 @@ class Auth:
             # Check if the Blockfrost API Key is configured in a environmental variable 
             assert (os.getenv('BLOCKFROST_API_KEY') is not None), '[ERROR] Your blockfrost api key is not configured in your environement path.'
             self.api_key = os.getenv('BLOCKFROST_API_KEY')
-        
         else:
             self.api_key = api_key
         
@@ -83,8 +43,7 @@ class Auth:
         response = query_blockfrost(url_stake_info, self.api_key, self.proxies)
         
         return response
-            
-            
+                 
     def stake_reward_history(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain the reward history.
@@ -106,7 +65,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-
     def stake_amount_history(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain the stake amount history.
@@ -128,10 +86,7 @@ class Auth:
         #print('[INFO] Function stake_amount_history, {} API calls.'.format(count_api_calls))
         
         return pd.DataFrame.from_dict(response) if pandas else response
-        
-       
-
-    
+         
     def stake_delegation(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain information about the stake delegation.
@@ -153,8 +108,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
-    
     def stake_registration_deregistrations(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain information about the stake registration et deregistrations.
@@ -176,7 +129,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def stake_withdrawal_history(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain information about the stake withdrawal history.
@@ -201,7 +153,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def stake_mir_history(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain information about the stake mir history.
@@ -223,7 +174,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def stake_associated_addresses(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain information about the stake associated addresses.
@@ -245,7 +195,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def stake_assets_associated_addresses(self, stake_address: str, data_order: str='asc', nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain information about the stake assets associated addresses.
@@ -267,7 +216,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def address_info(self, address: str) -> dict:
         """
         Obtain informations about an address.
@@ -283,8 +231,6 @@ class Auth:
         
         return response
     
-    
-    
     def address_details(self, address: str) -> dict:
         """
         Obtain details information about an address.
@@ -299,7 +245,6 @@ class Auth:
         
         return response
     
-    
     def address_utxo(self, address: str, pandas: bool=False) -> dict:
         """
         Obtain UTXO of the address.
@@ -314,7 +259,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def address_transaction(self, address: str, pandas: bool=False) -> dict:
         """
         Transactions on the address.
@@ -329,7 +273,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
     
-    
     def network_info(self) -> dict:
         """Return detailed network information."""
         
@@ -339,7 +282,6 @@ class Auth:
         
         return response
     
-    
     def latest_epoch(self) -> dict:
         """Return the information about the latest, therefore current, epoch."""
         
@@ -349,7 +291,6 @@ class Auth:
         
         return response
     
-    
     def specific_epoch(self, epoch: int) -> dict:
         """
         Obtain information about a specific epocht.
@@ -367,7 +308,6 @@ class Auth:
         
         return response
     
-    
     def latest_epoch_protocol_parameters(self) -> dict:
         """Return the protocol parameters for the latest epoch."""
         
@@ -376,8 +316,7 @@ class Auth:
         response = query_blockfrost(address_info_url, self.api_key, self.proxies)
         
         return response
-    
-    
+     
     def epochs_history(self, epochs: list, pandas: bool=False) -> Union[pd.DataFrame, dict]:            
         """
         Obtain history about the epochs.
@@ -409,7 +348,6 @@ class Auth:
             
         return pd.DataFrame.from_dict(epochs_history) if pandas else epochs_history
     
-    
     def registered_polls(self, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain the list of registered stake pools.
@@ -429,8 +367,7 @@ class Auth:
         #print('[INFO] Function registered_polls, {} API calls.'.format(count_api_calls))
                         
         return pd.DataFrame.from_dict(response) if pandas else response
-    
-    
+     
     def pool_informations(self, pool_id: str) -> dict: 
         """
         Obtain Pool information.
@@ -443,8 +380,7 @@ class Auth:
         response = query_blockfrost(pool_informations_url, self.api_key, self.proxies)
         
         return response
-    
-    
+      
     def stake_pool_history(self, pool_id: str, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain history of stake pool parameters over epochs
@@ -465,8 +401,7 @@ class Auth:
         #print('[INFO] Function param_stake_pool_history, {} API calls.'.format(count_api_calls))
         
         return pd.DataFrame.from_dict(response) if pandas else response
-
-    
+ 
     def rewards_history_analysis(self, stake_address: str, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Create a dataframe with explanatory variables of the stake rewards for each epochs.
@@ -520,12 +455,11 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
 
-
     def specific_asset(self, asset: str) -> dict: 
         """
         Obtain information about a specific asset.
         
-        :param : :param assets: Assets (Concatenation of the policy_id and hex-encoded asset_name)
+        :param assets: Assets (Concatenation of the policy_id and hex-encoded asset_name)
         :return: Dictionary with the info about the asset
         """
         
@@ -534,8 +468,7 @@ class Auth:
         response = query_blockfrost(specific_asset_url, self.api_key, self.proxies)
         
         return response
-
-        
+ 
     def asset_history(self, asset: str, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         Obtain the history of a specific asset.
@@ -555,7 +488,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
 
-
     def asset_transactions(self, asset: str, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         List of a specific asset transactions.
@@ -574,8 +506,7 @@ class Auth:
         #print('[INFO] Function asset_transactions, {} API calls.'.format(count_api_calls))
         
         return pd.DataFrame.from_dict(response) if pandas else response
-
-        
+     
     def asset_addresses(self, asset: str, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         List of a addresses containing a specific asset.
@@ -594,7 +525,6 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
 
-    
     def assets_policy(self, policy_id: str, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
         """
         List of asset minted under a specific policy.
@@ -613,8 +543,7 @@ class Auth:
         
         return pd.DataFrame.from_dict(response) if pandas else response
 
-
-    def assets_policy_informations(self, policy_id: str, nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, list]:
+    def assets_policy_info(self, policy_id: str, nb_of_results: int=None, pandas: bool=False) -> Union[pd.DataFrame, list]:
         '''
         Obtain informations about the assets minted under a specific policy ID.
 
@@ -648,31 +577,212 @@ class Auth:
 
         return (pd.DataFrame.from_dict(assets_data), assets_not_found)  if pandas else (assets_data, assets_not_found)
 
+    def specific_tx(self, txs_hash: str) -> dict: 
+        """
+        Obtain the content of the requested transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about the content of the transaction
+        """
+        
+        specific_tx_url = self.network + bf_tx_url + txs_hash
+
+        response = query_blockfrost(specific_tx_url, self.api_key, self.proxies)
+        
+        return response
+
+    def tx_utxos(self, txs_hash: str) -> dict: 
+        """
+        Return the inputs and UTXOs of the specific transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about the inputs and UTXOs of the specific transaction
+        """
+        
+        specific_tx_utxos_url = self.network + bf_tx_url + txs_hash + bf_tx_utxos_url
+
+        response = query_blockfrost(specific_tx_utxos_url, self.api_key, self.proxies)
+        
+        return response
+
+    def tx_stake_address_cert(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain information about (de)registration of stake addresses within a transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about (de)registration of stake addresses within a transaction
+        """
+        
+        stake_address_certificates_url = self.network + bf_tx_url + txs_hash + bf_stake_address_certificates_url
+
+        response = query_blockfrost(stake_address_certificates_url, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_delegation_cert(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain information about delegation certificates of a specific transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about delegation certificates of a specific transaction
+        """
+        
+        tx_delegation_certificates_url = self.network + bf_tx_url + txs_hash + bf_tx_delegation_certificates_url
+
+        response = query_blockfrost(tx_delegation_certificates_url, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_withdrawal_url(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain information about withdrawals of a specific transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about withdrawals of a specific transaction
+        """
+        
+        tx_withdrawal_url = self.network + bf_tx_url + txs_hash + bf_tx_withdrawal_url
+
+        response = query_blockfrost(tx_withdrawal_url, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_transaction_mirs(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain information about withdrawals of a specific transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about withdrawals of a specific transaction
+        """
+        
+        tx_transaction_mirs = self.network + bf_tx_url + txs_hash + bf_tx_transaction_mirs_url
+
+        response = query_blockfrost(tx_transaction_mirs, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_stake_pool_update(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain information about stake pool registration and update certificates of a specific transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about stake pool registration and update certificates of a specific transaction
+        """
+        
+        tx_stake_pool_update_url = self.network + bf_tx_url + txs_hash + bf_tx_stake_pool_update_url
+
+        response = query_blockfrost(tx_stake_pool_update_url, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_stake_pool_retirement_cert(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain information about stake pool retirements within a specific transaction.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about stake pool retirements within a specific transaction
+        """
+        
+        tx_stake_pool_retirement_cert_url = self.network + bf_tx_url + txs_hash + bf_tx_stake_pool_retirement_cert_url
+
+        response = query_blockfrost(tx_stake_pool_retirement_cert_url , self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_metadata(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain the transaction metadata.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about the transaction metadata
+        """
+        
+        tx_metadata_url = self.network + bf_tx_url + txs_hash + bf_tx_metadata_url
+
+        response = query_blockfrost(tx_metadata_url , self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_cbor_metadata(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain the transaction metadata in CBOR.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about the transaction metadata in CBOR
+        """
+        
+        tx_cbor_metadata_url = self.network + bf_tx_url + txs_hash + bf_tx_cbor_metadata_url
+
+        response = query_blockfrost(tx_cbor_metadata_url, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def tx_redeemers(self, txs_hash: str, pandas: bool=False) -> dict: 
+        """
+        Obtain the transaction redeemers.
+        
+        :param txs_hash: Transaction hash
+        :return: Dictionary with the info about the transaction redeemers.
+        """
+        
+        tx_redeemers_url = self.network + bf_tx_url + txs_hash + bf_tx_redeemers_url
+
+        response = query_blockfrost(tx_redeemers_url, self.api_key, self.proxies)
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def scripts_list(self, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
+        """
+        List of scripts.
+        
+        :param data_order: Optional, use 'desc' if you want reverse the data order (default: Ascending)
+        :param nb_of_results: Optional, Number of results wanted. The api return 100 results at a time (default: 100)
+        :param pandas: Optional, True for return a pandas dataframe (default: False)
+        :return: Dictionary or DataFrame of the scripts hash
+        """
+        
+        scripts_url = bf_assets_url + bf_scripts_url
+        response, count_api_calls = query_on_several_pages(self.network, self.api_key, data_order, nb_of_results, bf_scripts_url, self.proxies)
+        
+        #print('[INFO] Function script_list, {} API calls.'.format(count_api_calls))
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
+
+    def specific_script(self, script_hash: str) -> dict: 
+        """
+        Information about a specific script.
+        
+        :param script_hash: Script hash
+        :return: Dictionary with the info about a specific script.
+        """
+        
+        specific_script_url = self.network + bf_specific_script_url + script_hash
+
+        response = query_blockfrost(specific_script_url, self.api_key, self.proxies)
+        
+        return response
+
+    def redeem_specific_script(self, script_hash: str, data_order: str='asc', nb_of_results: int=100, pandas: bool=False) -> Union[pd.DataFrame, dict]:
+        """
+        List of redeemers of a specific script.
+        
+        :param script_hash: Script hash
+        :param data_order: Optional, use 'desc' if you want reverse the data order (default: Ascending)
+        :param nb_of_results: Optional, Number of results wanted. The api return 100 results at a time (default: 100)
+        :param pandas: Optional, True for return a pandas dataframe (default: False)
+        :return: Dictionary or DataFrame of the redeemers of a specific script
+        """
+        
+        redeem_specific_script_url = bf_specific_script_url + script_hash + bf_redeem_specific_script_url
+        response, count_api_calls = query_on_several_pages(self.network, self.api_key, data_order, nb_of_results, redeem_specific_script_url, self.proxies)
+        
+        #print('[INFO] Function redeem_specific_script, {} API calls.'.format(count_api_calls))
+        
+        return pd.DataFrame.from_dict(response) if pandas else response
       
-def set_query_string_parameter(page: int, data_order: str="") -> str:
-    """
-    Create the query string to add at the end of the request url.
-    
-    :param page: Page number to extract the data
-    :param data_order: Order of the data 
-    :return: Query string parameter
-    """
-   
-    # Data is returned in ascending (oldest first, newest last) order.
-    # Use the ?order=desc query parameter to reverse this order
-    order_parameter = lambda order: 'order=desc'if 'desc' in order else ''
-    
-    # By default, the api return 100 results at a time. You have to use ?page=2 to list through the results
-    page_parameter = 'page={}'.format(page)
-    
-    query_string_parameter = '?' + page_parameter + '&' + order_parameter(data_order)
-    
-    return query_string_parameter
-
-
 def query_blockfrost(url: str, api_key: str, proxies: dict=None) -> dict:
     """
-    Query blockfrost api.
+    Query Blockfrost API.
     
     :param url: The url
     :param api_key: Blockfrost api Key
@@ -695,32 +805,6 @@ def query_blockfrost(url: str, api_key: str, proxies: dict=None) -> dict:
                                                    url,
                                                    json['message']))
 
-
-def nb_results_to_return(nb_of_results: int) -> Tuple[int, bool]:
-    """
-    Set the variables for determinated the number of data to return.
-    If nb_of_results is None, nb_last_page is set to 0 for get all the data available.
-    
-    :param nb_of_results: Number of data to return, None for get all the data available
-    :return: The number of the last page to get the data (or 0 for get all the data) and False (True for get all the data) or raise an error
-    """
-
-    # Check if the nb_of_results is not equal to 0 or is not a multiple of 100
-    if nb_of_results != None:
-        assert(nb_of_results!=0), "[ERROR] The function parameter 'nb_of_results' ({}), cant be zero, he should be 100 or a multiple of 100 or None for get all the data available.".format(nb_of_results)
-        assert(nb_of_results%100 == 0), "[ERROR] The function parameter 'nb_of_results' ({}), should be 100 or a multiple of 100 or None for get all the data available.".format(nb_of_results)
-   
-    # Determine the number of the last page to get the data, O for request all the data available.
-    nb_last_page = lambda number: 0 if not number else int(number / 100) 
-
-    get_all_data = True
-    
-    if not nb_of_results:
-        return nb_last_page(nb_of_results), get_all_data
-
-    return nb_last_page(nb_of_results), False
-
-
 def query_on_several_pages(network: str, api_key: str, data_order: str, nb_of_results: int, query_url: str, proxies: dict) -> Tuple[dict, int]:
     """
     Get the data from several pages and the number of api calls.
@@ -774,7 +858,6 @@ def query_on_several_pages(network: str, api_key: str, data_order: str, nb_of_re
     return _dict, count_api_calls
 
 
-def process_onchain_metadata(asset: dict) -> Dict:
     """
     Create a column for each onchain metadata item.
     
@@ -797,7 +880,3 @@ def process_onchain_metadata(asset: dict) -> Dict:
         f_assets_data[asset_item] = asset[asset_item]
     
     return f_assets_data
-
-def convert_hex_to_ascii(hex_string: str) -> str:
-    """Convert hex string to ascii format"""
-    return bytearray.fromhex(hex_string).decode()
\ No newline at end of file
